/*
 * %CopyrightBegin%
 * 
 * Copyright Ericsson AB 1997-2021. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 * %CopyrightEnd%
 */
/*
 * System dependent driver declarations
 */

#ifndef __DRIVER_INT_H__
#define __DRIVER_INT_H__

#ifdef HAVE_SYS_UIO_H
#include <sys/types.h>
#include <sys/uio.h>

typedef struct iovec SysIOVec;

#else

typedef struct {
    char* iov_base;
    int   iov_len;
} SysIOVec;

#endif

#endif
