%%% This file was automatically generated by snmpc_mib_to_hrl
-ifndef('SNMP-USM-AES-MIB').
-define('SNMP-USM-AES-MIB', true).

%% Oids

-define(usmAesCfb128Protocol, [1,3,6,1,6,3,10,1,2,4]).

-define(snmpUsmAesMIB, [1,3,6,1,6,3,20]).


%% Range values


%% Default values

-endif.
