%%% This file was automatically generated by snmpc_mib_to_hrl
-ifndef('SNMPv2-TM').
-define('SNMPv2-TM', true).

%% Oids

-define(snmpUDPDomain, [1,3,6,1,6,1,1]).

-define(snmpCLNSDomain, [1,3,6,1,6,1,2]).

-define(snmpCONSDomain, [1,3,6,1,6,1,3]).

-define(snmpDDPDomain, [1,3,6,1,6,1,4]).

-define(snmpIPXDomain, [1,3,6,1,6,1,5]).

-define(rfc1157Proxy, [1,3,6,1,6,2,1]).

-define(rfc1157Domain, [1,3,6,1,6,2,1,1]).

-define(snmpv2TM, [1,3,6,1,6,3,0]).


%% Range values


%% Default values

-endif.
