%%
%% %CopyrightBegin%
%% 
%% Copyright Ericsson AB 2007-2024. All Rights Reserved.
%% 
%% Licensed under the Apache License, Version 2.0 (the "License");
%% you may not use this file except in compliance with the License.
%% You may obtain a copy of the License at
%%
%%     http://www.apache.org/licenses/LICENSE-2.0
%%
%% Unless required by applicable law or agreed to in writing, software
%% distributed under the License is distributed on an "AS IS" BASIS,
%% WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
%% See the License for the specific language governing permissions and
%% limitations under the License.
%% 
%% %CopyrightEnd%
%%

%%
%%----------------------------------------------------------------------
%% Purpose: Megaco erlang dist compress behaviour module
%%----------------------------------------------------------------------

-module(megaco_edist_compress).
-moduledoc """
Megaco erlang dist compress behaviour.

The following functions should be exported from a `megaco_edist_compress`
callback module:
""".

-doc """
Compress a megaco component. The erlang dist encoder makes no assumption on the
how or even if the component is compressed.
""".
-callback encode(R, Version) -> T when
      R       :: megaco_encoder:megaco_message() |
                 megaco_encoder:transaction() | 
                 megaco_encoder:action_reply() | 
                 megaco_encoder:action_request() | 
                 megaco_encoder:command_request(),
      Version :: megaco_encoder:protocol_version(),
      T       :: term().

-doc "Decompress a megaco component.".
-callback decode(T, Version) -> R when
      T       :: term(),
      Version :: megaco_encoder:protocol_version() | dynamic,
      R       :: megaco_encoder:megaco_message() |
                 megaco_encoder:transaction() | 
                 megaco_encoder:action_reply() | 
                 megaco_encoder:action_request() | 
                 megaco_encoder:command_request().

