%%% This file was automatically generated by snmpc_mib_to_hrl
-ifndef('SNMP-USM-HMAC-SHA2-MIB').
-define('SNMP-USM-HMAC-SHA2-MIB', true).

%% Oids

-define(snmpUsmHmacSha2MIB, [1,3,6,1,2,1,235]).

-define(usmHMAC128SHA224AuthProtocol, [1,3,6,1,6,3,10,1,1,4]).

-define(usmHMAC192SHA256AuthProtocol, [1,3,6,1,6,3,10,1,1,5]).

-define(usmHMAC256SHA384AuthProtocol, [1,3,6,1,6,3,10,1,1,6]).

-define(usmHMAC384SHA512AuthProtocol, [1,3,6,1,6,3,10,1,1,7]).


%% Range values


%% Default values

-endif.
