%%% This file was automatically generated by snmpc_mib_to_hrl
-ifndef('OTP-TC').
-define('OTP-TC', true).

%% Oids

-define(otpTcModule, [1,3,6,1,4,1,193,19,1,1,2]).


%% Range values


%% Default values

-endif.
