body {
  font: 14px/1.6 "Helvetica Neue", Helvetica, Arial, sans-serif;
  margin: 0;
  padding: 0;
  color: #000;
  border-top: 2px solid #ddd;
  background-color: #fff;

  min-height: 100%;
  display: flex;
  flex-direction: column;
}

h1 {
  width: 100%;
  border-bottom: 1px solid #eee;
  margin-bottom: 0;
  font-weight: 100;
  font-size: 1.1em;
  letter-spacing: 1px;
}

h1 code {
  font-size: 0.96em;
}

code {
  font: 12px monospace;
}

footer {
  background: #eee;
  width: 100%;
  padding: 10px 0;
  text-align: right;
  border-top: 1px solid #ddd;
  display: flex;
  flex: 1;
  order: 2;
  justify-content: center;
}

table {
  width: 100%;
  margin-top: 10px;
  border-collapse: collapse;
  border: 1px solid #cbcbcb;
  color: #000;
  -webkit-border-radius: 3px;
  -moz-border-radius: 3px;
}
table thead {
  display: none;
}
table td.line,
table td.line a,
table td.hits {
  width: 20px;
  background: #eaeaea;
  text-align: center;
  text-decoration: none;
  font-size: 11px;
  padding: 0 10px;
  color: #949494;
}
table td.hits {
  width: 10px;
  text-align: right;
  padding: 2px 5px;
  color: rgba(0, 0, 0, 0.6);
  background-color: #f0f0f0;
}
tr.miss td.line,
tr.miss td.line a,
tr.miss td.hits {
  background-color: #ffdce0;
  border-color: #fdaeb7;
}
tr.miss td {
  background-color: #ffeef0;
}
tr.hit td.line,
tr.hit td.line a,
tr.hit td.hits {
  background-color: #cdffd8;
  border-color: #bef5cb;
}
tr.hit td {
  background-color: #e6ffed;
}
td.source {
  padding-left: 15px;
  line-height: 15px;
  white-space: pre;
  font: 12px monospace;
}
